-----------------------------------------------------------------------
-- Locale
--

local l = GetLocale()

local TOO_MANY = "Too many results, %d not shown. Whisper me !next to get more."
local PREFIX = "cg: "
local CG_USAGE = "Whisper me '!%s <anything>' to search."
local NO_DATA = "No results found for %q."
local NO_SKILL = "I don't know this skill. Accepted commands are %s."
local ACCEPTED = "Accepted commands are %s."
local HELP = {
	"CraftyGnome accepts various !<skill> commands in whispers or guild chat based on what trade skills the player has available.",
	"Usage: !command <token1> <token2> ...",
	"Each token is matched against all the text that is present in all the recipe tooltips available to this player.",
	"Only tooltips that contain all the given tokens are returned as results to you.",
}

-- All tradeskills ninjaed from Babble-Tradeskill-2.2
local skills = nil
if l == "enUS" then
	skills = {
		["Alchemy"] = true,
		["Blacksmithing"] = true,
		["Cooking"] = true,
		["Enchanting"] = true,
		["Engineering"] = true,
		["First Aid"] = true,
		["Jewelcrafting"] = true,
		["Leatherworking"] = true,
		["Tailoring"] = true,
	}
	for k, v in pairs(skills) do
		skills[k] = k
	end
elseif l == "deDE" then
	skills = {
		["Alchemy"] = "Alchimie",
		["Blacksmithing"] = "Schmiedekunst",
		["Cooking"] = "Kochkunst",
		["Enchanting"] = "Verzauberkunst",
		["Engineering"] = "Ingenieurskunst",
		["First Aid"] = "Erste Hilfe",
		["Jewelcrafting"] = "Juwelenschleifen",
		["Leatherworking"] = "Lederverarbeitung",
		["Tailoring"] = "Schneiderei",
	}
elseif l == "frFR" then
	skills = {
		["Alchemy"] = "Alchimie",
		["Blacksmithing"] = "Forge",
		["Cooking"] = "Cuisine",
		["Enchanting"] = "Enchantement",
		["Engineering"] = "Ingnierie",
		["First Aid"] = "Premiers soins",
		["Jewelcrafting"] = "Joaillerie",
		["Leatherworking"] = "Travail du cuir",
		["Tailoring"] = "Couture",
	}
elseif l == "zhCN" then
	skills = {
		["Alchemy"] = "炼金术",
		["Blacksmithing"] = "锻造",
		["Cooking"] = "烹饪",
		["Enchanting"] = "附魔",
		["Engineering"] = "工程学",
		["First Aid"] = "急救",
		["Jewelcrafting"] = "珠宝加工",	--TBC check
		["Leatherworking"] = "制皮",
		["Tailoring"] = "裁缝",
	}
elseif l == "zhTW" then
	skills = {
		["Alchemy"] = "鍊金術",
		["Blacksmithing"] = "鍛造",
		["Cooking"] = "烹飪",
		["Enchanting"] = "附魔",
		["Engineering"] = "工程學",
		["First Aid"] = "急救",
		["Jewelcrafting"] = "珠寶設計",
		["Leatherworking"] = "製皮",
		["Tailoring"] = "裁縫",
	}
elseif l == "koKR" then
	skills = {
		["Alchemy"] = "연금술",
		["Blacksmithing"] = "대장기술",
		["Cooking"] = "요리",
		["Enchanting"] = "마법부여",
		["Engineering"] = "기계공학",
		["First Aid"] = "응급치료",
		["Jewelcrafting"] = "보석세공",
		["Leatherworking"] = "가죽세공",
		["Tailoring"] = "재봉술",
	}
elseif l == "esES" then
	skills = {
		["Alchemy"] = "Alquimia",
		["Blacksmithing"] = "Herreria",
		["Cooking"] = "Cocina",
		["Enchanting"] = "Encantamiento",
		["Engineering"] = "Ingenieria",
		["First Aid"] = "Primeros auxilios",
		["Jewelcrafting"] = "Joyeria",
		["Leatherworking"] = "Peleteria",
		["Tailoring"] = "Sastreria",
	}
end

-----------------------------------------------------------------------
-- Tooltip scanner, mostly stolen from Gratuity.
--

local enchantLink = "|cffffd000|Henchant:%d|h[Placeholder Name]|h|r"
local scanItem
do
	local scanner = CreateFrame("GameTooltip")
	scanner:SetOwner(WorldFrame, "ANCHOR_NONE")
	local lcache, rcache = {}, {}
	for i = 1, 30 do
		lcache[i], rcache[i] = scanner:CreateFontString(), scanner:CreateFontString()
		lcache[i]:SetFontObject(GameFontNormal); rcache[i]:SetFontObject(GameFontNormal)
		scanner:AddFontStrings(lcache[i], rcache[i])
	end
	local function erase()
		for i = 1, scanner:NumLines() do rcache[i]:SetText() end
		scanner:ClearLines()
	end
	local scannerCache = {}
	function scanItem(id)
		erase()
		local link = enchantLink:format(id)
		scanner:SetHyperlink(link)
		for k in pairs(scannerCache) do scannerCache[k] = nil end
		local name = nil
		for i = 1, scanner:NumLines() do
			local left, right = lcache[i]:GetText(), rcache[i]:GetText()
			if not name then name = left end
			if left then table.insert(scannerCache, left) end
			if right then table.insert(scannerCache, right) end
		end
		return name, table.concat(scannerCache)
	end
end

-----------------------------------------------------------------------
-- CraftyGnome
--

local bangMappings = {
	alchemy = skills["Alchemy"],
	blacksmith = skills["Blacksmithing"],
	cooking = skills["Cooking"],
	leather = skills["Leatherworking"],
	enchanting = skills["Enchanting"],
	engineering = skills["Engineering"],
	firstaid = skills["First Aid"],
	jewel = skills["Jewelcrafting"],
	tailor = skills["Tailoring"],
}
bangMappings.alc = bangMappings.alchemy
bangMappings.bs = bangMappings.blacksmith
bangMappings.cook = bangMappings.cooking
bangMappings.lw = bangMappings.leather
bangMappings.enchant = bangMappings.enchanting
bangMappings.eng = bangMappings.engineering
bangMappings.aid = bangMappings.firstaid
bangMappings.jc = bangMappings.jewel
bangMappings.gem = bangMappings.jewel
bangMappings.tailoring = bangMappings.tailor
bangMappings.craft = true
bangMappings.cg = true
bangMappings.next = true
bangMappings.help = true

local buildRecipeLink = "|cffffd000|Henchant:%d|h[%s]|h|r"
local MAX_RESULTS = 5
local db = nil
local cache = {}
local idToName = {} -- Can be removed in 2.4
local linkToId = {} -- Can be removed in 2.4
local acceptedCommands = {}
local addon = CreateFrame("Frame")

-- we can use GetSpellLink and GetSpellInfo on the recipe IDs in 2.4, so remove
-- idToName, linkToId and the link building and such alltogether.
--
-- XXX perhaps we should switch to a PT-like storage scheme, like
-- "recipeId1:lvl recipeId2:lvl recipeId3:lvl"

function addon:CRAFT_SHOW()
	local name = GetCraftName()
	if not name then return end
	if not db[name] then db[name] = {} end
	for i = 1, GetNumCrafts() do
		local skillName, _, skillType, _, _, _, level = GetCraftInfo(i)
		if skillType ~= "header" and skillName then
			local recipeId = select(3, GetCraftRecipeLink(i):find("enchant:(%d+)"))
			recipeId = tonumber(recipeId)
			local l = level or 1
			if not db[name][recipeId] or db[name][recipeId] < l then
				db[name][recipeId] = l
			end
		end
	end
end

function addon:TRADE_SKILL_SHOW()
	local name = GetTradeSkillLine()
	if not name then return end
	if not db[name] then db[name] = {} end
	for i = 1, GetNumTradeSkills() do
		local skillName, skillType = GetTradeSkillInfo(i)
		if skillType ~= "header" and skillName then
			local _, _, _, level = GetItemInfo(GetTradeSkillItemLink(i))
			local recipeId = select(3, GetTradeSkillRecipeLink(i):find("enchant:(%d+)"))
			recipeId = tonumber(recipeId)
			local l = level or 1
			if not db[name][recipeId] or db[name][recipeId] < l then
				db[name][recipeId] = l
			end
		end
	end
end

do
	local ctl = ChatThrottleLib
	local reply = ctl and
		function(msg, player) ctl:SendChatMessage("BULK", "CraftyGnome", PREFIX .. msg, "WHISPER", nil, player) end or
		function(msg, player) SendChatMessage(PREFIX .. msg, "WHISPER", nil, player) end

	local function updateCache(skill)
		if not cache[skill] then cache[skill] = {} end
		for id in pairs(db[skill]) do
			if not cache[skill][id] then
				local name, details = scanItem(id)
				idToName[id] = name
				cache[skill][id] = details:lower()
			end
		end
	end

	local function updateAcceptedCommands()
		for k in pairs(acceptedCommands) do acceptedCommands[k] = nil end
		for k, v in pairs(bangMappings) do
			if db[v] then
				table.insert(acceptedCommands, "!" .. k)
			end
		end
		table.insert(acceptedCommands, "!help")
		table.insert(acceptedCommands, "!craft")
		table.insert(acceptedCommands, "!cg")
	end

	local function search(skill, ...)
		local tbl = nil
		if cache[skill] then
			local c = select("#", ...)
			for id, details in pairs(cache[skill]) do
				local matchAll = true
				for i = 1, c do
					local token = tostring(select(i, ...)):lower()
					if not details:find(token) then
						matchAll = nil
					end
				end
				if matchAll then
					if not tbl then tbl = {} end
					local link = buildRecipeLink:format(id, idToName[id])
					linkToId[link] = id
					table.insert(tbl, link)
				end
			end
		end
		return tbl
	end

	local moreResults = {}
	local function replyResults(tbl, player)
		if not tbl then return end
		if #tbl > MAX_RESULTS then
			for i = 1, MAX_RESULTS do
				if tbl[i] then reply(tbl[i], player) end
			end
			for i = 1, MAX_RESULTS do table.remove(tbl, 1) end
			reply(TOO_MANY:format(#tbl), player)
		else
			for i, v in ipairs(tbl) do reply(v, player) end
			tbl = nil
		end
		return tbl
	end

	local bangHandlers = {}
	bangHandlers.next = function(player)
		moreResults[player] = replyResults(moreResults[player], player)
	end
	bangHandlers.craft = function(player)
		updateAcceptedCommands()
		reply(ACCEPTED:format(table.concat(acceptedCommands, ", ")), player)
	end
	bangHandlers.cg = bangHandlers.craft
	bangHandlers.help = function(player, isWhisper)
		if not isWhisper then return end
		for i, v in pairs(HELP) do
			reply(v, player)
		end
		bangHandlers.craft(msg, player)
	end

	local skillSorters = {}
	local function processChat(msg, player, isWhisper)
		local cmd = select(3, msg:find("^!(%S+)"))
		if not cmd or not bangMappings[cmd] then return end
		if bangHandlers[cmd] then
			bangHandlers[cmd](player, isWhisper)
			return
		end
		local query = select(3, msg:find("^!%S+%s(.*)"))
		if query then
			local skill = bangMappings[cmd]
			if db[skill] then
				updateCache(skill)

				local name = GetItemInfo(query)
				local results = nil
				if name then
					-- This is an item, don't split it up into multiple tokens
					results = search(skill, name)
				else
					results = search(skill, strsplit(" ", query))
				end
				if results then
					if not skillSorters[skill] then
						skillSorters[skill] = function(a, b)
							return db[skill][linkToId[a]] > db[skill][linkToId[b]]
						end
					end
					table.sort(results, skillSorters[skill])
					moreResults[player] = replyResults(results, player)
				elseif isWhisper then
					reply(NO_DATA:format(query), player)
				end
			else
				updateAcceptedCommands()
				reply(NO_SKILL:format(table.concat(acceptedCommands, ", ")), player)
			end
		elseif isWhisper then
			reply(CG_USAGE:format(cmd), player)
		end
	end
	function addon:CHAT_MSG_GUILD(msg, player)
		processChat(msg, player)
	end
	function addon:CHAT_MSG_WHISPER(msg, player)
		processChat(msg, player, true)
	end
end

function addon:PLAYER_ENTERING_WORLD()
	if not CraftyGnomeDB then CraftyGnomeDB = {} end
	local realm = GetRealmName()
	if not CraftyGnomeDB[realm] then CraftyGnomeDB[realm] = {} end
	local faction = UnitFactionGroup("player")
	if not CraftyGnomeDB[realm][faction] then CraftyGnomeDB[realm][faction] = {} end
	db = CraftyGnomeDB[realm][faction]

	if not db.version or db.version < 4 then
		for k in pairs(db) do db[k] = nil end
		db.version = 4
	end

	addon:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

local orig = ChatFrame_MessageEventHandler
local chatPrefixTest = "^" .. PREFIX
ChatFrame_MessageEventHandler = function(event, ...)
	if event == "CHAT_MSG_WHISPER" then
		local cmd = select(3, arg1:find("^!(%S+)"))
		if cmd and bangMappings[cmd] then return end
	elseif event == "CHAT_MSG_WHISPER_INFORM" and arg1:find(chatPrefixTest) then
		return
	end
	return orig(event, ...)
end

addon:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
addon:RegisterEvent("PLAYER_ENTERING_WORLD")
addon:RegisterEvent("CHAT_MSG_GUILD")
addon:RegisterEvent("CHAT_MSG_WHISPER")
addon:RegisterEvent("CRAFT_SHOW")
addon:RegisterEvent("TRADE_SKILL_SHOW")

